/* c5.c - Don Yang (uguu.org)

   Remove pointers.

   10/13/01
                                                                           */

#include<stdio.h>
#include<stdlib.h>

#define INTERNAL_NODE   256
#define UNUSED_NODE     300

typedef struct _node
{
   int x, y;
   int left, right, next;
} Node;

Node HeapPool[512];
unsigned char buffer[1024];
int Dictionary[256], Heap, a, b, c;

int HeapPoolIndex, Tree[2][256], branch,
   f[257], NodeCount,
   FileSize, size, i, j, bit, byte;

FILE *infile, *outfile;

/**************************************************************** AssignID */
static void AssignID(int node, int parent)
{
   HeapPool[node].next = parent;
   if( HeapPool[node].y != INTERNAL_NODE )
   {
      Dictionary[HeapPool[node].y] = node;
      return;
   }
   HeapPool[node].x = NodeCount++;
   AssignID(HeapPool[node].left, node);
   AssignID(HeapPool[node].right, node);

   Tree[0][HeapPool[node].x] =
      (HeapPool[HeapPool[node].left].y == INTERNAL_NODE)
         ?  HeapPool[HeapPool[node].left].x
         :  -HeapPool[HeapPool[node].left].y;
   Tree[1][HeapPool[node].x] =
      (HeapPool[HeapPool[node].right].y == INTERNAL_NODE)
         ?  HeapPool[HeapPool[node].right].x
         :  -HeapPool[HeapPool[node].right].y;

} /* AssignID() */

/**************************************************************** HeapPush */
static void HeapPush(void)
{
   c = HeapPoolIndex++;

   HeapPool[c].x = f[i];
   HeapPool[c].y = i;
   HeapPool[c].left = a;
   HeapPool[c].right = b;

   if( Heap < 0 || HeapPool[Heap].x > HeapPool[c].x )
   {
      HeapPool[c].next = Heap;
      Heap = c;
   }
   else
   {
      a = Heap;
      b = HeapPool[a].next;
      while( b != -1 && HeapPool[b].x < HeapPool[c].x )
      {
         a = b;
         b = HeapPool[b].next;
      }
      HeapPool[a].next = c;
      HeapPool[c].next = b;
   }

} /* HeapPush() */

/*************************************************************** WriteBits */
static void WriteBits(int node)
{
   if( HeapPool[node].next != -1 )
   {
      WriteBits(HeapPool[node].next);
      byte |= (HeapPool[HeapPool[node].next].right == node ? 1 : 0) << bit++;
      if( bit > 7 )
      {
         fputc(byte, outfile);
         bit = byte = 0;
      }
   }

} /* WriteBits() */

/******************************************************************** main */
int main(int argc, char **argv)
{
   if( argc < 3 )
      return printf("%s <infile> <outfile>\n", *argv);

   if( (infile = fopen(argv[1], "rb")) == NULL )
      return printf("Can not open %s\n", argv[1]);
   if( (outfile = fopen(argv[2], "wb+")) == NULL )
   {
      fclose(infile);
      return printf("Can not create %s\n", argv[2]);
   }

   fseek(infile, 0, SEEK_SET);
   size = fread(buffer, 1, 579, infile);
   if( size == 0 )
   {
      fclose(infile);
      fclose(outfile);
      return puts("Empty file");
   }

   /* LoadHeader() */
   NodeCount = buffer[0];
   FileSize = 0;
   if( NodeCount * 2 + ((NodeCount + 3) / 4) + 5 <= size )
   {
      /* LoadTree() */
      for(i = 0; i < NodeCount; i += 4)
      {
         for(j = 0; j < 4; j++)
            f[i + j] = ((buffer[i / 4 + 1] >> (j * 2)) & 3);
      }

      j = (NodeCount + 3) / 4 + 1;
      for(i = 0; i < NodeCount; i++)
      {
         for(c = 0; c < 2; c++)
         {
            if( (f[i] & (c + 1)) != 0 )
            {
               Tree[c][i] = -buffer[j++];
            }
            else
            {
               Tree[c][i] = buffer[j++];
               if( Tree[c][i] >= NodeCount || Tree[c][i] <= i )
                  break;
            }
         }
         if( c < 2 )
            break;
      }
      if( i == NodeCount )
         FileSize = (buffer[j] | (buffer[j + 1] << 8) |
                    (buffer[j + 2] << 16) | (buffer[j + 3] << 24));
   }
   if( FileSize > 0 )
   {
      /* DecodeFile() */
      fseek(infile, NodeCount * 2 + (NodeCount + 3) / 4 + 5, SEEK_SET);
      byte = j = 0;
      bit = 8;
      for(i = 0; i < FileSize;)
      {
         if( bit > 7 )
         {
            byte = fgetc(infile);
            bit = 0;
         }

         branch = (byte & (1 << bit)) == 0 ? 0 : 1;
         if( Tree[branch][j] <= 0 )
         {
            fputc(-Tree[branch][j], outfile);
            j = 0;
            i++;
         }
         else
         {
            j = Tree[branch][j];
         }
         bit++;
      }
      printf("Decoded %s -> %s\n", argv[1], argv[2]);
   }
   else
   {
      /* CreateTree() */
      for(i = 0; i < 256; f[i++] = 0);

      FileSize = 0;
      fseek(infile, 0, SEEK_SET);
      while( (size = fread(buffer, 1, 1024, infile)) != 0 )
      {
         for(i = 0; i < size; f[buffer[i++]]++);
         FileSize += size;
      }

      HeapPoolIndex = 0;
      Heap = -1;
      a = b = 0;

      for(i = 0; i < 256; i++)
      {
         if( f[i] != 0 )
            HeapPush();
      }
      if( HeapPool[Heap].next == -1 )
      {
         for(i = 0; f[i] > 0; i++);
         HeapPush();
      }

      i = INTERNAL_NODE;
      do
      {
         /* HeapPop() */
         a = Heap;

         /* HeapPop() */
         b = HeapPool[Heap].next;
         Heap = HeapPool[HeapPool[Heap].next].next;

         f[i] = HeapPool[a].x + HeapPool[b].x;
         HeapPush();
      } while( HeapPool[Heap].next != -1 );

      NodeCount = 0;
      AssignID(Heap, -1);

      for(i = NodeCount; i < 256; i++)
         Tree[0][i] = Tree[1][i] = UNUSED_NODE;

      /* WriteHeader() */
      fputc(NodeCount, outfile);
      for(i = 0; i < NodeCount; i += 4)
      {
         byte = 0;
         for(j = 0; j < 4; j++)
         {
            if( Tree[0][i + j] <= 0 )  byte |= 1 << (j * 2);
            if( Tree[1][i + j] <= 0 )  byte |= 2 << (j * 2);
         }
         fputc(byte, outfile);
      }

      for(i = 0; i < NodeCount; i++)
      {
         fputc(abs(Tree[0][i]), outfile);
         fputc(abs(Tree[1][i]), outfile);
      }

      for(i = 0; i < 4; i++)
         fputc((FileSize >> (i * 8)) & 255, outfile);

      /* EncodeFile() */
      bit = byte = 0;
      fseek(infile, 0, SEEK_SET);
      for(i = 0; i < FileSize; i++)
      {
         j = fgetc(infile);
         WriteBits(Dictionary[j]);
      }

      if( bit > 0 )
         fputc(byte, outfile);

      printf("Encoded %s -> %s\n", argv[1], argv[2]);
   }

   fclose(infile);
   fclose(outfile);
   return 0;

} /* main() */
